{% extends "!layout.html" %}
{% block footer %}
{{ super() }}
<script async src="https://www.googletagmanager.com/gtag/js?id=UA-134061901-2"></script>
<script>
  window.dataLayer = window.dataLayer || [];
  function gtag(){dataLayer.push(arguments);}
  gtag('js', new Date());
  gtag('config', 'UA-134061901-2');
</script>
{% endblock %}